package net.greenfieldtech.cloudonixsdk.cloudonixdialer.api;

import static org.xbill.DNS.Type.TXT;

import android.annotation.TargetApi;
import android.text.TextUtils;
import androidx.annotation.NonNull;

import com.google.gson.Gson;
import java.io.EOFException;
import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.Inet4Address;
import java.net.InetAddress;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import net.greenfieldtech.cloudonixsdk.cloudonixdialer.BuildConfig;
import net.greenfieldtech.cloudonixsdk.cloudonixdialer.api.request.RegistrationFreeSessionRequest;
import net.greenfieldtech.cloudonixsdk.cloudonixdialer.api.request.RegistrationFreeTokenRequest;
import net.greenfieldtech.cloudonixsdk.cloudonixdialer.api.response.AuthResponse;
import net.greenfieldtech.cloudonixsdk.cloudonixdialer.api.response.RegistrationFreeDialingResponse;
import net.greenfieldtech.cloudonixsdk.utils.managers.ConnectivityManager;
import okhttp3.Call;
import okhttp3.Callback;
import okhttp3.MediaType;
import okhttp3.OkHttpClient;
import okhttp3.Request;
import okhttp3.RequestBody;
import okhttp3.Response;
import okhttp3.logging.HttpLoggingInterceptor;
import org.xbill.DNS.Lookup;
import org.xbill.DNS.Record;
import org.xbill.DNS.SimpleResolver;
import org.xbill.DNS.TXTRecord;

/**
 * AppRestClient.java
 * 
 * Authors:
 *   - Alexander Rubanskiy, 2019-10-03
 *   - Oded Arbel, 2019-11-20
 * 
 * Copyright © 2015 GreenfieldTech 2015 Ltd. All rights reserved.
 * Copyright © 2018 Cloudonix Ltd. All rights reserved.
 * Copyright © 2019 Cloudonix Inc. All rights reserved.
 */
public class AppRestClient {

  private static final MediaType JSON = MediaType.parse("application/json; charset=utf-8");
  private static final String DEFAULT_BASE_API_URL = "https://api.cloudonix.io";

  private static OkHttpClient getRestClient() {
    return new OkHttpClient.Builder()
        .addInterceptor(new HttpLoggingInterceptor().setLevel(getLogLevel()))
        .addNetworkInterceptor(new AppRequestInterceptor())
        .connectTimeout(30, TimeUnit.SECONDS)
        .readTimeout(30, TimeUnit.SECONDS)
        .writeTimeout(30, TimeUnit.SECONDS)
        .build();
  }

  private static HttpLoggingInterceptor.Level getLogLevel() {
    return BuildConfig.DEBUG ? HttpLoggingInterceptor.Level.BODY
        : HttpLoggingInterceptor.Level.NONE;
  }

  private static ExecutorService dnsResolver = Executors.newSingleThreadExecutor();

  public static void createRegistrationFreeSession(String callee, String msisdn, String pass,
      String sipServer, String domain, RegistrationFreeSessionCallback callback) {
    String baseApi = getBaseApi(sipServer, domain);

    auth(baseApi, msisdn, pass, domain, new AuthCallback() {
      @Override
      public void onAuthSuccess(String domain, String msisdn, String token) {
        getRegFreeSession(baseApi, callee, domain, msisdn, token, callback);
      }

      @Override
      public void onAuthFailure() {
        callback.onSessionFailure();
      }
    });
  }

  private static void auth(String baseApi, String msisdn, String pass, String domain,
      AuthCallback callback) {
    RegistrationFreeTokenRequest request = new RegistrationFreeTokenRequest(pass);
    RequestBody body = RequestBody.create(JSON, new Gson().toJson(request));
    Request requestBody = new Request.Builder()
        .url(baseApi + "/auth/subscribers/" + msisdn + "@" + domain)
        .post(body)
        .build();

    getRestClient().newCall(requestBody).enqueue(new Callback() {
      @Override
      public void onFailure(@NonNull Call call, @NonNull IOException e) {
        if (callback != null) {
          callback.onAuthFailure();
        }
      }

      @Override
      public void onResponse(@NonNull Call call, @NonNull Response response) throws IOException {
        if (response.code() / 100 != 2) {
          onFailure(call, new IOException("Received error response from Reg-Free session: " + response.code()));
          return;
        }
        AuthResponse tokenResponse = new Gson()
            .fromJson(response.body().string(), AuthResponse.class);
        if (TextUtils.isEmpty(tokenResponse.getBearer())) {
          if (callback != null) {
            callback.onAuthFailure();
          }
        }
        if (callback != null) {
          callback.onAuthSuccess(domain, msisdn, tokenResponse.getBearer());
        }
      }
    });
  }

  private static void getRegFreeSession(String baseApi, String number, String domain, String msisdn,
      String token, RegistrationFreeSessionCallback callback) {
    RegistrationFreeSessionRequest request = new RegistrationFreeSessionRequest(number);
    RequestBody body = RequestBody.create(JSON, new Gson().toJson(request));
    Request requestBody = new Request.Builder()
        .url(baseApi + "/calls/" + domain + "/outgoing/" + msisdn)
        .addHeader("Authorization", token)
        .post(body)
        .build();
    getRestClient().newCall(requestBody).enqueue(new Callback() {

      @Override
      public void onFailure(@NonNull Call call, @NonNull IOException e) {
        if (callback != null) {
          callback.onSessionFailure();
        }
      }

      @Override
      public void onResponse(@NonNull Call call, @NonNull Response response) throws IOException {
        RegistrationFreeDialingResponse dialingResponse =
            new Gson().fromJson(response.body().string(), RegistrationFreeDialingResponse.class);
        if (callback != null) {
          callback.onSessionSuccess(dialingResponse);
        }
      }
    });
  }

  public static void getRingingNotification(String msisdn, String sipServer, String domain,
      String sessionId, RingingListener callback) {
    String baseApi = getBaseApi(sipServer, domain);

    Request requestBody = new Request.Builder()
        .url(String.format(baseApi + "/calls/%s/ringing/%s/%s", domain, msisdn, sessionId))
        .get()
        .build();

    getRestClient().newCall(requestBody).enqueue(new Callback() {
      @Override
      public void onFailure(@NonNull Call call, @NonNull IOException e) {
        if (e instanceof EOFException) {
          if (callback != null) {
            callback.onRingingEnd();
          }
        }
      }

      @Override
      public void onResponse(@NonNull Call call, @NonNull Response response) {
        if (response.code() == HttpURLConnection.HTTP_FORBIDDEN) {
          if (callback != null) {
            callback.onRingingEnd();
          }
        }
      }
    });
  }

  public static void deleteIncomingSession(String msisdn, String pass, String sipServer,
      String domain, String sessionId) {
    String baseApi = getBaseApi(sipServer, domain);

    auth(baseApi, msisdn, pass, domain, new AuthCallback() {
      @Override
      public void onAuthSuccess(String domain, String msisdn, String token) {
        deleteSession(baseApi, domain, sessionId, token);
      }

      @Override
      public void onAuthFailure() {

      }
    });
  }

  private static void deleteSession(String baseApi, String domain, String sessionId, String token) {
    Request requestBody = new Request.Builder()
        .url(String.format(baseApi + "/calls/%s/sessions/%s?reason=denied", domain, sessionId))
        .addHeader("Authorization", token)
        .delete()
        .build();

    getRestClient().newCall(requestBody).enqueue(new Callback() {
      @Override
      public void onFailure(@NonNull Call call, @NonNull IOException e) {

      }

      @Override
      public void onResponse(@NonNull Call call, @NonNull Response response) {

      }
    });
  }

  private static String getBaseApi(String sipServer, String domain) {
    String apiBase;
    if (sipServer.toLowerCase().endsWith(".cloudonix.io")) {
      apiBase = "api" + sipServer.substring(sipServer.indexOf("."));
    } else {
      apiBase = getCloudonixAPI(domain);
    }
    return !(apiBase.startsWith("http://")) && !(apiBase.startsWith("https://")) ? "https://"
        + apiBase : apiBase;
  }

  @TargetApi(24)
  private static String getCloudonixAPI(final String domain) {
    try {
      return dnsResolver.submit(() -> {
        Lookup lookup = new Lookup("_api." + domain, TXT);
        List<InetAddress> servers = ConnectivityManager.getInstance().getDnsServers();
        SimpleResolver resolver = new SimpleResolver();
        InetAddress selectedAdd = null;
        if (servers != null) {
          for (InetAddress address : servers) {
            if (address instanceof Inet4Address) {
              selectedAdd = address;
              break;
            }
            selectedAdd = servers.get(0);
          }
          resolver.setAddress(selectedAdd);
        }
        lookup.setResolver(resolver);
        Record[] records = lookup.run();
        for (Record rec : records) {
          if (!(rec instanceof TXTRecord)) {
            continue;
          }
          TXTRecord txt = (TXTRecord) rec;
          return (String) txt.getStrings().get(0);
        }
        return DEFAULT_BASE_API_URL;
      }).get();
    } catch (Exception e) {
      return DEFAULT_BASE_API_URL;
    }
  }

  public interface AuthCallback {
    void onAuthSuccess(String domain, String msisdn, String token);
    void onAuthFailure();
  }

  public interface RegistrationFreeSessionCallback {
    void onSessionSuccess(RegistrationFreeDialingResponse response);
    void onSessionFailure();
  }

  public interface RingingListener {
    void onRingingEnd();
  }

}
