package net.greenfieldtech.cloudonixsdk.cloudonixdialer.utils;

import android.annotation.TargetApi;
import android.text.TextUtils;

import com.google.gson.Gson;

import net.greenfieldtech.cloudonixsdk.api.rest.requests.BaseRequest;
import net.greenfieldtech.cloudonixsdk.cloudonixdialer.BuildConfig;
import net.greenfieldtech.cloudonixsdk.cloudonixdialer.models.registration_free_dialing.RegistrationFreeDialingResponse;
import net.greenfieldtech.cloudonixsdk.cloudonixdialer.models.registration_free_dialing.RegistrationFreeSessionRequest;
import net.greenfieldtech.cloudonixsdk.cloudonixdialer.models.registration_free_dialing.RegistrationFreeTokenRequest;
import net.greenfieldtech.cloudonixsdk.cloudonixdialer.models.registration_free_dialing.RegistrationFreeTokenResponse;
import net.greenfieldtech.cloudonixsdk.utils.SDKLogger;
import net.greenfieldtech.cloudonixsdk.utils.managers.ConnectivityManager;

import java.io.IOException;
import java.net.Inet4Address;
import java.net.InetAddress;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import org.xbill.DNS.Lookup;
import org.xbill.DNS.Record;
import org.xbill.DNS.SimpleResolver;
import org.xbill.DNS.TXTRecord;
import org.xbill.DNS.TextParseException;

import okhttp3.Call;
import okhttp3.Callback;
import okhttp3.Interceptor;
import okhttp3.MediaType;
import okhttp3.OkHttpClient;
import okhttp3.Request;
import okhttp3.RequestBody;
import okhttp3.Response;
import okhttp3.logging.HttpLoggingInterceptor;

import static org.xbill.DNS.Type.TXT;

/**
 * RegistrationFreeClient.java
 * 
 * Authors:
 *   - Alexander Rubanskiy, 2017-06-19
 *   - Oded Arbel, 2019-11-13
 * 
 * Copyright © 2015 GreenfieldTech 2015 Ltd. All rights reserved.
 * Copyright © 2018 Cloudonix Ltd. All rights reserved.
 * Copyright © 2019 Cloudonix Inc. All rights reserved.
 */
public class RegistrationFreeClient {
	
	public static final MediaType JSON = MediaType.parse("application/json; charset=utf-8");
	public static final String DEFAULT_BASE_API_URL = "https://api.cloudonix.io";
	public static final String TAG = "DialerRestClient";
	
	private static HttpLoggingInterceptor.Level getLogLevel() {
		return BuildConfig.DEBUG ? HttpLoggingInterceptor.Level.BODY
				: HttpLoggingInterceptor.Level.NONE;
	}
	
	private static OkHttpClient getRestClient() {
		return new OkHttpClient.Builder()
				.addInterceptor(new HttpLoggingInterceptor().setLevel(getLogLevel()))
				.addNetworkInterceptor(new Interceptor() {
					@Override
					public Response intercept(Chain chain) throws IOException {
						Request originalRequest = chain.request();
						Request newRequest = originalRequest.newBuilder().header("User-Agent",
								"Cloudonix Dialer/5.2").build();
						return chain.proceed(newRequest);
					}
				})
				.connectTimeout(30, TimeUnit.SECONDS)
				.readTimeout(30, TimeUnit.SECONDS)
				.writeTimeout(30, TimeUnit.SECONDS)
				.build();
	}
	
	public static void createSession(String number, String caller, String pass, String domain,
									 String serverName, SessionCallback callback) {
		String apiBase;
		if (serverName != null && serverName.endsWith(".cloudonix.io")) {
			String[] parts = serverName.split("\\.");
			apiBase = "api";
			for (int i = parts.length - 3; i < parts.length; ++i)
				apiBase += "." + parts[i];
		} else {
			apiBase = getCloudonixAPI(domain);
		}
		final String apiUrl = !(apiBase.startsWith("http://")) && !(apiBase.startsWith("https://"))?  "https://" + apiBase : apiBase;
		SDKLogger.i(TAG, "Resolve API server to " + apiUrl);
		
		BaseRequest request = new RegistrationFreeTokenRequest(pass);
		RequestBody body = RequestBody.create(JSON, new Gson().toJson(request));
		Request requestBody = new Request.Builder()
				.url(apiUrl + "/auth/subscribers/" + caller + "@" + domain)
				.post(body)
				.build();
		
		getRestClient().newCall(requestBody).enqueue(new Callback() {
			@Override
			public void onFailure(Call call, IOException e) {
				callback.onFailure(e);
			}
			
			@Override
			public void onResponse(Call call, Response response) throws IOException {
				RegistrationFreeTokenResponse tokenResponse = new Gson()
						.fromJson(response.body().string(), RegistrationFreeTokenResponse.class);
				
				if (TextUtils.isEmpty(tokenResponse.getBearer())) {
					callback.onFailure(new IOException(tokenResponse.getMessage()));
					return;
				}
				
				BaseRequest request = new RegistrationFreeSessionRequest(number);
				RequestBody body = RequestBody.create(JSON, new Gson().toJson(request));
				Request requestBody = new Request.Builder()
						.url(apiUrl + "/calls/" + domain + "/outgoing/" + caller)
						.addHeader("Authorization", tokenResponse.getBearer())
						.post(body)
						.build();
				
				getRestClient().newCall(requestBody).enqueue(new Callback() {
					@Override
					public void onFailure(Call call, IOException e) {
						callback.onFailure(e);
					}
					
					@Override
					public void onResponse(Call call, Response response) throws IOException {
						RegistrationFreeDialingResponse dialingResponse =
								new Gson().fromJson(response.body().string(), RegistrationFreeDialingResponse.class);
						callback.onSuccess(dialingResponse);
					}
				});
			}
		});
	}
	
	
	private static ExecutorService dnsResolver = Executors.newSingleThreadExecutor();
	
	private static String getCloudonixAPI(final String domain) {
		try {
			return dnsResolver.submit(new Callable<String>() {
				@Override
				public String call() throws Exception {
					Lookup lookup = new Lookup("_api." + domain, TXT);
					List<InetAddress> servers = ConnectivityManager.getInstance().getDnsServers();
					SimpleResolver resolver = new SimpleResolver();
					InetAddress selectedAdd = null;
					if (servers != null) {
						for (InetAddress address: servers) {
							if(address instanceof Inet4Address){
								selectedAdd  = address;
								break;
							}
							selectedAdd = servers.get(0);
						}
						resolver.setAddress(selectedAdd);
					}
					lookup.setResolver(resolver);
					Record[] records = lookup.run();
					for (Record rec : records) {
						if (!(rec instanceof TXTRecord))
							continue;
						TXTRecord txt = (TXTRecord) rec;
						return (String)txt.getStrings().get(0);
					}
					return DEFAULT_BASE_API_URL;
				}
			}).get();
		} catch (Exception e) {
			return DEFAULT_BASE_API_URL;
		}
	}
	
	public interface SessionCallback {
		void onSuccess(RegistrationFreeDialingResponse response);
		void onFailure(IOException e);
	}
}
