package net.greenfieldtech.cloudonixsdk.cloudonixdialer.utils;

import android.Manifest;
import android.app.Activity;
import android.content.Context;
import android.content.Intent;
import android.content.pm.PackageManager;
import android.util.Log;
import android.widget.Toast;
import androidx.core.app.ActivityCompat;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.List;
import net.greenfieldtech.cloudonixsdk.api.interfaces.IVoIPObserver;
import net.greenfieldtech.cloudonixsdk.api.interfaces.IVoIPObserver.CallState;
import net.greenfieldtech.cloudonixsdk.api.interfaces.IVoIPObserver.ConfigurationKey;
import net.greenfieldtech.cloudonixsdk.api.models.RegistrationData;
import net.greenfieldtech.cloudonixsdk.api.models.SDKConstants;
import net.greenfieldtech.cloudonixsdk.api.models.StatisticsData;
import net.greenfieldtech.cloudonixsdk.appinterface.CloudonixSDKClient;
import net.greenfieldtech.cloudonixsdk.appinterface.DefaultVoipObserver;
import net.greenfieldtech.cloudonixsdk.appinterface.SIPHeader;
import net.greenfieldtech.cloudonixsdk.appinterface.VoipCall;
import net.greenfieldtech.cloudonixsdk.cloudonixdialer.R;
import net.greenfieldtech.cloudonixsdk.cloudonixdialer.activities.CallActivity;
import net.greenfieldtech.cloudonixsdk.cloudonixdialer.api.AppRestClient;
import net.greenfieldtech.cloudonixsdk.cloudonixdialer.api.response.RegistrationFreeDialingResponse;
import net.greenfieldtech.cloudonixsdk.cloudonixdialer.utils.prefs.AppPrefs;
import net.greenfieldtech.cloudonixsdk.cloudonixdialer.utils.prefs.AppPrefsKeys;
import net.greenfieldtech.cloudonixsdk.utils.SDKLogger;

/**
 * VoIPClient.java
 * 
 * Authors:
 *   - Aleksei Dzis, 2016-08-03
 * 
 * Copyright © 2015 GreenfieldTech 2015 Ltd. All rights reserved.
 * Copyright © 2018 Cloudonix Ltd. All rights reserved.
 * Copyright © 2019 Cloudonix Inc. All rights reserved.
 */
public class VoIPClient {

  private static final String TAG = "VoIPClient";

  public static final String INTENT_CALLING_ACTION = "net.greenfieldtech.cloudonixsdk.CALL_ACTIVE";
  public static final String INTENT_RINGING_ACTION = "net.greenfieldtech.cloudonixsdk.CALL_RINGING";
  public static final String INTENT_EARLY_MEDIA_ACTION = "net.greenfieldtech.cloudonixsdk.CALL_EARLY_MEDIA";
  public static final String INTENT_CALL_HANGUP = "net.greenfieldtech.cloudonixsdk.CALL_HANGUP";
  public static final String INTENT_INCOMING_CALL = "net.greenfieldtech.cloudonixsdk.CALL_INCOMING";

  private static String LICENSE_KEY;

  private static VoIPClient mInstance;
  private CloudonixSDKClient sdkClient;

  private boolean isSpeakerOn = false;

  public static VoIPClient getInstance() {
    return mInstance;
  }

  private CloudonixSDKClient getWrapper() {
    if (sdkClient == null) {
      sdkClient = CloudonixSDKClient.getInstance(LICENSE_KEY);
    }
    return sdkClient;
  }

  private VoIPClient(Context context) {
    getWrapper().setLogLevel(SDKLogger.LOG_LEVEL_ALL);
    getWrapper().bind(context);

    SIPConfig.create(context);

    getWrapper().setAllowIncomingCallIntent(INTENT_INCOMING_CALL);

    Intent intent = new Intent(context, CallActivity.class);
    intent.setAction("net.greenfieldtech.sipsdk.ACTIVE_CALL");
    getWrapper().setNotificationResources(intent, R.mipmap.ic_launcher,
        context.getString(R.string.app_name), "SDK is running and has an active call");
    getWrapper().setForegroundInitRes(R.mipmap.ic_launcher,
        "VoIP Service","Initializing","VoIP Service initializing");
    getWrapper().addEventsListener(new DefaultVoipObserver(){
      @Override
      public void onLog(int level, String message) {
        if (!message.contains("[MobileSDK"))
          android.util.Log.println(level, "VoIPClient", message);
      }
    });
  }

  public void destroy() {
    getWrapper().removeFromForeground();
    getWrapper().destroy();
    mInstance = null;
  }

  public void shutdown() {
    getWrapper().shutdown();
  }

  public void restartSip() {
    getWrapper().restart();
  }

  public void setConfig(ConfigurationKey key, String value) {
    getWrapper().setConfig(key, value);
  }

  public void commitConfig() {
    getWrapper().commitConfig();
  }

  public StatisticsData getStatistics(String key) {
    return getWrapper().getStatistics(key);
  }

  // First init must be with internet connection !!!
  public static void initWithContext(Context context) {
    LICENSE_KEY = loadLicenseKey(context);
    mInstance = new VoIPClient(context);
  }

  private static String loadLicenseKey(Context context) {
    InputStream input = context.getResources().openRawResource(R.raw.cloudonix_license_key);
    try {
      byte[] inputArray = convertStreamToByteArray(input);
      return new String(inputArray);
    } catch (IOException e) {
      e.printStackTrace();
    }
    return "";
  }

  public void addCallbacksListener(IVoIPObserver callbacksListener) {
    getWrapper().addEventsListener(callbacksListener);
  }

  public void removeCallbacksListener(IVoIPObserver listener) {
    getWrapper().removeEventsListener(listener);
  }

  public void setMediaResources(Context context,
      int busyTone,
      int ringingTone,
      int alertTone,
      int endCallTone,
      int holdTone,
      int beginHoldTone,
      int endHoldTone) {
    getWrapper().setMediaResources(context, busyTone, ringingTone, alertTone, endCallTone, holdTone,
        beginHoldTone, endHoldTone);
  }

  public boolean isRegistered() {
    return getWrapper().isRegistered();
  }

  public boolean isSpeakerOn() {
    return isSpeakerOn;
  }

  public void setSpeaker(boolean speakerEnabled) {
    isSpeakerOn = speakerEnabled;
    getWrapper().setAudioRoute(isSpeakerOn ?
        SDKConstants.AudioRoute.SPEAKER : SDKConstants.AudioRoute.EARPIECE);
  }

  public String getAudioRoute() {
    return getWrapper().getAudioRoute().toString();
  }

  public boolean isBluetoothOn() {
    return getWrapper().isBluetoothEnabled();
  }

  public void setBluetooth(boolean bluetoothEnabled) {
    if (bluetoothEnabled) {
      getWrapper().setAudioRoute(SDKConstants.AudioRoute.BLUETOOTH);
    } else {
      getWrapper().setAudioRoute(SDKConstants.AudioRoute.EARPIECE);
    }
  }

  public void setMute(boolean shouldMute) {
    getWrapper().mute(shouldMute);
  }

  public void setMuteByKey(String key, boolean shouldMute) {
    getWrapper().muteByKey(key, shouldMute);
  }

  public void sendDTMF(String key, char c) {
    getWrapper().DTMF(key, c);
  }

  public boolean isStarted() {
    return getWrapper().isInitialized();
  }

  public void answer(String callKey) {
    getWrapper().answer(callKey);
  }

  public void reject(String callKey) {
    getWrapper().reject(callKey);
  }

  public boolean dial(String number) {
    SDKLogger.d(TAG, "VoIPClient -> dial: " + number);
    if (!getWrapper().isRegistered()) {
      return false;
    }
    getWrapper().dial(number);
    return false;
  }

  public boolean dialWithHeaders(String number, List<SIPHeader> headers) {
    SDKLogger.d(TAG, "VoIPClient -> dial: ");
    if (!getWrapper().isRegistered()) {
      return false;
    }
    getWrapper().dialWithHeaders(number, headers);
    return false;
  }

  public void dialRegistrationFree(final String number) {
    SDKLogger.d(TAG, "VoIPClient -> dial free: " + number);

    AppPrefs prefs = AppPrefs.getInstance();
    final String userName = prefs.getString(AppPrefsKeys.USER_NAME);
    final String userPass = prefs.getString(AppPrefsKeys.PASSWORD);
    final String sipServer = prefs.getString(AppPrefsKeys.SERVER);
    final String domain = prefs.getString(AppPrefsKeys.DOMAIN);

    AppRestClient.createRegistrationFreeSession(number, userName, userPass, sipServer, domain,
        new AppRestClient.RegistrationFreeSessionCallback() {

          @Override
          public void onSessionSuccess(RegistrationFreeDialingResponse response) {
            dialRegistrationFree(number, response.getToken());
          }
  
          @Override
          public void onSessionFailure() {

          }
        });
  }

  public void dialRegistrationFree(final String number, String sessionId) {
    getWrapper().dialRegistrationFree(number, sessionId);
  }

  public void notifyRegistrationFreeIncoming(String callerId, String sessionId) {
    Log.d(TAG,"Caller id:"+callerId+" , and session id is: "+sessionId);
    getWrapper().notifyRegistrationFreeIncoming(callerId, sessionId);
  }

  public void setControlAudioRoute(boolean controlAudioRoute) {
    getWrapper().setControlAudioRoute(controlAudioRoute);
  }

  public boolean isControlAudioRoute() {
    return getWrapper().isControlAudioRoute();
  }

  public void unregister() {
    SDKLogger.d(TAG, "VoIPClient -> unregister()");
    getWrapper().unregister();
  }

  public float getCallMOS(String key) {
    StatisticsData data = getWrapper().getStatistics(key);
    if (data != null) {
      return data.MOS;
    } else {
      return 0;
    }
  }

  public void hangup(String callKey) {
    getWrapper().hangup(callKey);
  }

  public void localHold(String callKey, boolean enabled) {
    getWrapper().localHold(callKey, enabled);
  }

  public void dumpStatistics(boolean b) {
    getWrapper().dumpStatisticsToLog(b);
  }

  public void setEcFixedDelay(int fixedEcDelay) {
    getWrapper().setEcFixedDelay(fixedEcDelay);
  }

  public CallState getCallState(String callKey) {
    return getWrapper().getCallState(callKey);
  }

  public String getCallUrl(String key) {
    return getWrapper().getCallUrl(key);
  }

  public int getNumberOfCalls() {
    return getWrapper().getNumberOfCalls();
  }

  public String getCallKeyForIndex(int index) {
    return getWrapper().getCallKeyForIndex(index);
  }

  public VoipCall[] getCalls() {
    return getWrapper().getCalls();
  }

  public void removeForegroundService() {
    getWrapper().removeFromForeground();
  }

  public void addForegroundService() {
    // Call this only after setting notification resources via:
    // setNotificationResources()
    getWrapper().addToForeground();
  }

  public void askForPermissions(Activity activity, boolean isFirstTime) {
    String[] MANDATORY_PERMISSIONS = { Manifest.permission.RECORD_AUDIO };

    if (isFirstTime) {
      ActivityCompat.requestPermissions(activity, MANDATORY_PERMISSIONS, 100);
    } else {
      for (String permission : MANDATORY_PERMISSIONS) {
        if (activity.checkCallingOrSelfPermission(permission)
            != PackageManager.PERMISSION_GRANTED) {
          Toast.makeText(activity, "Permission " + permission
                  + " is not granted. Please accept it for the app to work correctly.",
              Toast.LENGTH_SHORT).show();
          ActivityCompat.requestPermissions(activity, new String[]{permission}, 100);
          return;
        }
      }
    }
  }

  public void setConfigs() {
    SIPConfig config = SIPConfig.getInstance();
    setConfig(ConfigurationKey.USER_AGENT, config.getUserAgent());
    setConfig(ConfigurationKey.USE_OPUS, config.getUseOpus() ? "1" : "0");
    setConfig(ConfigurationKey.USE_G711, config.getUseG711() ? "1" : "0");
    setConfig(ConfigurationKey.USE_G722, config.getUseG722() ? "1" : "0");
    setConfig(ConfigurationKey.USE_ILBC, config.getUseILBC() ? "1" : "0");
    setConfig(ConfigurationKey.MIN_RATE, config.getMinAudioRate());
    setConfig(ConfigurationKey.MAX_RATE, config.getMaxAudioRate());

    setConfig(ConfigurationKey.ENABLE_ICE, config.getEnableIce() ? "1" : "0");
    setConfig(ConfigurationKey.ENABLE_NAT, "1");
    setConfig(ConfigurationKey.DISABLE_SECURE_SIPS, "1");
//    setConfig(ConfigurationKey.LOG_LEVEL, "0");
    commitConfig();
  }

  public void setStoredConfiguration() {
    setConfiguration(getStoredConfiguration());
  }

  public void setConfiguration(RegistrationData registrationData) {
    getWrapper().setConfiguration(registrationData);
  }

  public String getSDKVersion() {
    return getWrapper().getSDKVersion();
  }

  private RegistrationData getStoredConfiguration() {
    AppPrefs prefs = AppPrefs.getInstance();

    RegistrationData registrationData = new RegistrationData();
    registrationData.setServerUrl(prefs.getString(AppPrefsKeys.SERVER));
    registrationData.setUsername(prefs.getString(AppPrefsKeys.USER_NAME));
    registrationData.setDisplayName(prefs.getString(AppPrefsKeys.DISPLAY_NAME));
    registrationData.setPassword(prefs.getString(AppPrefsKeys.PASSWORD));
    String portStr = prefs.getString(AppPrefsKeys.TRANSPORT_PORT);
    int port = 0;
    try {
      port = Integer.parseInt(portStr);
      if(!(port > 0 && port < 65536)){
        throw new NumberFormatException();
      }
    } catch (NumberFormatException e) {
      SDKLogger.e(TAG, "Port field is not valid: " + portStr);
      Log.e(TAG, "The Log says Port field is not valid: " + portStr);
    }

    registrationData.setPort(port);
    registrationData.setUseDnsSrv(prefs.getBoolean(AppPrefsKeys.USE_DNS_SRV));
    registrationData.setDomain(prefs.getString(AppPrefsKeys.DOMAIN));

    String transportType = prefs.getString(AppPrefsKeys.TRANSPORT);
    try {
      registrationData.setTransportType(RegistrationData.TransportType.valueOf(transportType));
    } catch (IllegalArgumentException e) {
      registrationData.setTransportType(RegistrationData.TransportType.TRANSPORT_TYPE_UDP);
    }

    registrationData.setWorkflowType(RegistrationData.WorkflowType.WORKFLOW_TYPE_AUTODETECT);

    return registrationData;
  }

  public void registerAccount() {
    getWrapper().registerAccount();
  }

  public void bind(Context context) {
    getWrapper().bind(context);
  }

  public void unbind() {
    getWrapper().unbind();
  }

	private static byte[] convertStreamToByteArray(InputStream is) throws IOException {
		ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
		byte[] buff = new byte[1024];
		int i;
		while ((i = is.read(buff, 0, buff.length)) > 0) {
			outputStream.write(buff, 0, i);
		}
		return outputStream.toByteArray();
	}
}
